//--------------------------------------------------------------------------//
// xmtnimage31.cc                                                           //
// fits and pds files                                                       //
// Latest revision: 06-12-2000                                              //
// See xmtnimage.h for Copyright Notice                                     //
//--------------------------------------------------------------------------//

#include "xmtnimage.h"

extern Globals     g;
extern Image      *z;
extern int         ci;
extern PrinterStruct printer;


//--------------------------------------------------------------------------//
//  readpdsfile                                                             //
//--------------------------------------------------------------------------//
int readpdsfile(char *filename)
{  
    return(readpds2file(filename));
}            

//--------------------------------------------------------------------------//
//  readpds2file                                                            //
//  Comments embedded within a label cause remainder of label to be ignored.//
//--------------------------------------------------------------------------//
int readpds2file(char *filename)
{  
  int format, status=OK, bpp=8, xsize=0, ysize=0, index=0, len=0,
      bytesperpixel, colortype=GRAY, xpos=0, ypos=0,x,x2,y,
      suffix_bytes=0, record_bytes=0, image_ptr=0,
      label_records=0;
  uchar cval=0;
  uint image_start=0;
  float fval=0;
  double dval=0;
  FILE *fp;
  if ((fp=fopen(filename,"r")) == NULL)
  {    error_message(filename, errno);
       status=NOTFOUND;
       return(status);
  }

  char tempstring[1024];
  char record[2880];
  char name[128];
  char value[128];
  char buf[10];
  char *equals=0;

  while(1)
  {   fgets(record,2880,fp); 
      if(record[0]!='/' || record[1]!='*')
      {    memcpy(name, record, 20);
           name[20]=0;
           if(name[0]=='E' && name[1]=='N' && name[2]=='D' && name[3]<=32) break;
           equals = 2+strchr(record,'=');
           len = strlen(record) - (equals - record);
           memcpy(value, equals, len);
           value[min(len-2, (int)strcspn(value, "/*"))] = 0;
           if(!strcmp(name,"RECORD_BYTES        ")) record_bytes=atoi(value);
           //// if(!strcmp(name,"FILE_RECORDS        ")) file_records=atoi(value);
           if(!strcmp(name,"LABEL_RECORDS       ")) label_records=atoi(value);
           //// if(!strcmp(name,"TRAILER_RECORDS     ")) trailer_records=atoi(value);
           if(!strcmp(name,"^IMAGE              ")) image_ptr=atoi(value);
           if(!strcmp(name,"IMAGE_LINES         ")) ysize=atoi(value);
           if(!strcmp(name,"LINE_SAMPLES        ")) xsize=atoi(value);
           if(!strcmp(name,"LINE_SUFFIX_BYTES   ")) suffix_bytes=atoi(value);
           if(!strcmp(name,"SAMPLE_BITS         ")) bpp=atoi(value);
           //// if(!strcmp(name,"SAMPLE_BIT_MASK     ")) bitmask=atoi(value);
     }
  }  
  

  bytesperpixel = (7+bpp)/8;
  if(image_ptr) image_start = record_bytes * image_ptr;
  else image_start = record_bytes * label_records;
  fseek(fp, image_start, SEEK_SET);

  if(record_bytes > xsize*bytesperpixel && suffix_bytes==0) 
     suffix_bytes = xsize*bytesperpixel - record_bytes;

  ////  Check parameters
  if(xsize<=0 || ysize<=0)
  {   message("Illegal image size", ERROR);
      fclose(fp);
      return ERROR;
  }

  format=INTEGER;
  switch(bpp)
  {   case 8:  format=CHARACTER; break;
      case 16: format=WORD; break;
      case 32: format=INTEGER; break;
      case -32: bpp=24; format=FLOAT; break;
      case -64: bpp=24; format=DOUBLE; break;
      default: 
          sprintf(tempstring,"Unsupported image depth: %d",bpp);
          message(tempstring, ERROR);
          fclose(fp);
          return ERROR;
  }

  if(newimage(basefilename(filename),
              xpos, ypos,       
              xsize, ysize,
              bpp, colortype,
              1,       // frames
              g.want_shell,
              1,       // scrollbars
              PERM, 0,
              g.window_border,
              0)!=OK) { fclose(fp); return(NOMEM); }

  char *buffer = new char[record_bytes];
  for(y=0;y<ysize;y++)  
  {   fread(buffer, record_bytes, 1, fp);
      index = 0;
      for(x=0,x2=0;x<xsize;x++,x2+=bytesperpixel)  
      {   switch(format)
          {   case CHARACTER: cval = buffer[index++];
                              z[ci].image[0][y][x] = cval; 
                              break;
              case WORD:      z[ci].image[0][y][x2+1]=buffer[index++]-128;
                              z[ci].image[0][y][x2]=buffer[index++]-128;
                              break;
              case INTEGER:   z[ci].image[0][y][x2+3]=buffer[index++]-128;
                              z[ci].image[0][y][x2+2]=buffer[index++]-128;
                              z[ci].image[0][y][x2+1]=buffer[index++]-128;
                              z[ci].image[0][y][x2+0]=buffer[index++]-128;
                              break;
              case FLOAT:     memcpy(buf, buffer+index, 4);
                              buf[4]=0;
                              fval = atof(buf);
                              index+=4;
                              putpixelbytes(z[ci].image[0][y]+x2,(int)fval,1,32,1);
                              break;
              case DOUBLE:    memcpy(buf, buffer+index, 8);
                              buf[8]=0;
                              dval = atof(buf);
                              index+=8;
                              putpixelbytes(z[ci].image[0][y]+x2,(int)dval,1,32,1);
                              break;
          }
      }
  }
  fclose(fp);
  setSVGApalette(1); 
  z[ci].colortype=GRAY;
  z[ci].hitgray=0;
  delete[] buffer;
  return 0;
}          


//--------------------------------------------------------------------------//
//  aboutpdsfile                                                            //
//--------------------------------------------------------------------------//
void aboutpdsfile(char *filename)
{  
  static listinfo *l;
  static char *listtitle;
  listtitle = new char[100];
  strcpy(listtitle, "About the File");
  static char **info;                                    
  int done=0, k; 
  FILE *fp;
  if ((fp=fopen(filename,"r")) == NULL)
  {    error_message(filename, errno);
       return;
  }

  char record[2880];
  int lc=-1;
  info = new char*[100];   // max. 100 lines - Change if you add more items

  info[++lc] = new char[100];                       
  sprintf(info[lc],"Filename:  %s",basefilename(filename));
  info[++lc]=new char[100]; sprintf(info[lc],"File format: PDS (Planetary Data System)");

  while(1)
  {   fgets(record,2880,fp); 
      if(record[0]=='E' && record[1]=='N' && record[2]=='D' && record[3]<=32) done=1;
      for(k=0;k<(int)strlen(record);k++) if(record[k]=='\n') record[k]=0;
      if(done || lc>=100) break;
      info[++lc]=new char[100]; strcpy(info[lc],record);
  }  
  
  l = new listinfo;
  l->title = listtitle;
  l->info  = info;
  l->count = lc+1;
  l->itemstoshow = min(20,lc+1);
  l->firstitem   = 0;
  l->wantsort    = 0;
  l->wantsave    = 0;
  l->helptopic   = 0;
  l->allowedit   = 0;
  l->selection   = &g.crud;
  l->width       = 0;
  l->transient   = 1;
  l->wantfunction = 0;
  l->autoupdate   = 0;
  l->deletebutton = 0;
  l->clearbutton  = 0;
  l->exportbutton = 0;
  l->highest_erased = 0;
  l->x = 0;
  l->y = 0;
  l->f1 = null;
  l->f2 = null;
  l->f3 = null;
  l->f4 = delete_list;
  l->f5 = null;
  l->f6 = null;
  l->listnumber = 0;
  list(l);
  return;
}  
          

//--------------------------------------------------------------------------//
//  aboutfitsfile                                                           //
//--------------------------------------------------------------------------//
void aboutfitsfile(char *filename)
{  
  static listinfo *l;
  static char *listtitle;
  listtitle = new char[100];
  strcpy(listtitle, "About the File");
  static char **info;                                    
  int done=0, k;
  FILE *fp;
  if ((fp=fopen(filename,"r")) == NULL)
  {    error_message(filename, errno);
       return;
  }

  char record[2880];
  char card[81];
  card[80]=0;
  int lc=-1;
  info = new char*[100];   // max. 100 lines - Change if you add more items

  info[++lc] = new char[100];                       
  sprintf(info[lc],"Filename:  %s",basefilename(filename));
  info[++lc]=new char[100]; sprintf(info[lc],"File format: FITS (Flexible Image Transport System)");

  while(!done)
  { 
      fread(&record,2880,1,fp); 
      for(k=0;k<36;k++)
      {    memcpy((void*)card, record+k*80, 80);
           if(card[0]=='E' && card[1]=='N' && card[2]=='D' && card[3]<=32) done=1;
           if(done || lc>=99) break;
           card[80] = 0;
           info[++lc]=new char[100]; strncpy(info[lc],card,99);
      }
      if(done || lc>=99) break;
  }
  fclose(fp);

  l = new listinfo;
  l->title = listtitle;
  l->info  = info;
  l->count = lc+1;
  l->itemstoshow = min(20,lc+1);
  l->firstitem   = 0;
  l->wantsort    = 0;
  l->wantsave    = 0;
  l->helptopic   = 0;
  l->allowedit   = 0;
  l->selection   = &g.crud;
  l->width       = 0;
  l->transient   = 1;
  l->wantfunction = 0;
  l->autoupdate   = 0;
  l->deletebutton = 0;
  l->clearbutton  = 0;
  l->exportbutton = 0;
  l->highest_erased = 0;
  l->x = 0;
  l->y = 0;
  l->f1 = null;
  l->f2 = null;
  l->f3 = null;
  l->f4 = delete_list;
  l->f5 = null;
  l->f6 = null;
  l->listnumber = 0;
  list(l);
  return;
}  


//--------------------------------------------------------------------------//
//  readfitsfile                                                            //
//--------------------------------------------------------------------------//
int readfitsfile(char *filename)
{  
  int done=0, format=INTEGER, i, status=OK, naxis=0, naxis3=0, k, bpp=8, bitpix=0, 
      xsize=0, ysize=0, hit=0, mcount=0,
      bytesperpixel, colortype=GRAY, xpos=0, ypos=0,x,x2,y;
  int rr;
  uchar cval=0;
  int answer;
  long int lival,filepos;
  int ival=0;
  float fval=0;
  double dval=0, datamax=65535.0, uscale=65535.0;
  FILE *fp;
  if ((fp=fopen(filename,"r")) == NULL)
  {    error_message(filename, errno);
       status=NOTFOUND;
       return(status);
  }

  char tempstring[1024];
  char record[2880];
  char card[81];
  char keyword[8];
  char value[80];
  keyword[7]=0;
  card[80]=0;
  uchar buf[32];
  int extend = 0;

  while(!done)
  { 
      fread(&record,2880,1,fp); //// Only 1 record at top of file
      for(k=0;k<36;k++)
      {    memcpy((void*)card, record+k*80, 80);
           memcpy(keyword, card, 7);
           memcpy(value,   card+9, 70);
           if(!strcmp(keyword,"END    ")) done=1;
           if(done) break;
           if(!strcmp(keyword,"BITPIX ")) bitpix=atoi(value);  // bits/channel
           if(!strcmp(keyword,"NAXIS  ")) naxis=atoi(value);
           if(!strcmp(keyword,"NAXIS1 ")) xsize=atoi(value);
           if(!strcmp(keyword,"NAXIS2 ")) ysize=atoi(value);           
           if(!strcmp(keyword,"NAXIS3 ")) naxis3=atoi(value);           
          // if(!strcmp(keyword,"DATAMIN")) datamin=atof(value);           
           if(!strcmp(keyword,"DATAMAX")) datamax=atof(value);           
           if(!strcmp(keyword,"USCALE ")) uscale=atof(value);   // not in the spec but apparently important        
           //// if(!strcmp(keyword,"BSCALE ")) bscale=atof(value);           
           //// if(!strcmp(keyword,"BZERO  ")) bzero=atof(value);           
           if(!strcmp(keyword,"EXTEND ")) extend++;  // have to read more junk at end
      }

      //// Now we hopefully have enough information to allocate image buffer
      if(naxis && xsize && ysize && !hit)   
      {
           hit = 1;
           ////  Check parameters
           if(naxis!=2 && naxis!=3)
           {     sprintf(tempstring, "%d axes in FITS file\nnot supported",naxis);
                 message(tempstring, ERROR);
                 fclose(fp);
                 return ERROR;
           }

           if(xsize<=0 || ysize<=0)
           {     message("Illegal image size", ERROR);
                 fclose(fp);
                 return ERROR;
           }


           switch(bitpix)
           {     case 8:   bpp=8;  format=CHARACTER; break;
                 case 16:  bpp=16; format=WORD; break;
                 case 32:  
                           bpp=16; 
                           format=INTEGER; 
                           message("Converting 32-bit integer\ndata to 16-bit integers");
                           break;
                 case -32: 
                           bpp=16; 
                           format=FLOAT; 
                           message("Converting 32-bit IEEE 754 floating point\ndata to 16-bit integers");
                           break;
                 case -64: 
                           bpp=16; 
                           format=DOUBLE; 
                           message("Converting 64-bit IEEE 754 floating point\ndata to 16-bit integers");
                           break;
                 default: 
                      sprintf(tempstring,"Unsupported image depth: %d",bpp);
                      message(tempstring, ERROR);
                      fclose(fp);
                      return ERROR;
           }
           switch(naxis)
           {     case 2: colortype=GRAY; break;
                 case 3: colortype=COLOR; bpp*=naxis3; break;  // hope it's a 3
                 default: colortype=GRAY;
           }
           bytesperpixel = (7+bpp)/8;

           if(newimage(basefilename(filename),
                       xpos, ypos,       
                       xsize, ysize,
                       bpp, colortype,
                       1,       // frames
                       g.want_shell,
                       1,       // scrollbars
                       PERM, 0,
                       g.window_border,
                       0)!=OK) { fclose(fp); return(NOMEM); }

           z[ci].metadata = new char*[MAXRECORDS];
      }

      z[ci].metadata[mcount] = new char[2880];
      strcpy(z[ci].metadata[mcount], record);
      mcount++;
  }  

  datamax = fmin(datamax,uscale);

  z[ci].subformat = format;
  if(colortype==GRAY)
  {    
       z[ci].colortype=GRAY;
       z[ci].hitgray=0;
       bytesperpixel = (7+z[ci].bpp)/8;
       for(y=ysize-1;y>=0;y--)  
       for(x=0,x2=0;x<xsize;x++,x2+=bytesperpixel)  
       {     switch(format)
             {     case CHARACTER: 
                                   fread(&cval,1,1,fp);
                                   z[ci].image[0][y][x] = cval; 
                                   break;
                   case WORD:      
                                   fread(&buf,2,1,fp);
     #ifdef LITTLE_ENDIAN
                                   z[ci].image[0][y][x2+1]=buf[0]-128;
                                   z[ci].image[0][y][x2]=buf[1];
     #else
                                   z[ci].image[0][y][x2]=buf[0]-128;
                                   z[ci].image[0][y][x2+1]=buf[1];
     #endif
                                   break;
                   case INTEGER:   fread(&ival,4,1,fp);
                                   putpixelbytes(z[ci].image[0][y]+x2,ival,1,32,1);
                                   break;
                   case FLOAT:     fread(&ival,4,1,fp);
     #ifdef LITTLE_ENDIAN
                                   swapbytes((uchar*)&ival,4,4);
     #endif
                                   fval = ieee754(ival);
                                   answer = min(65535, max(0,(int)(datamax*fval)));
                                   putpixelbytes(z[ci].image[0][y]+x2,answer,1,32,1);
                                   break; 
                   case DOUBLE:    fread(&lival,8,1,fp); 
     #ifdef LITTLE_ENDIAN
                                   swapbytes((uchar*)&lival,8,8);
     #endif
                                   dval = ieee754d(lival);
                                   answer = min(65535, max(0,(int)(datamax*dval)));
                                   putpixelbytes(z[ci].image[0][y]+x2,answer,1,32,1);
                                   break;
             }
       }
  }else
  {    for(i=0; i<3; i++)
       {     
             if(i==0) { g.wantr=1;g.wantg=0;g.wantb=0; }
             if(i==1) { g.wantr=0;g.wantg=1;g.wantb=0; }
             if(i==2) { g.wantr=0;g.wantg=0;g.wantb=1; }

             for(y=ysize-1; y>=0; y--)  
             for(x=0,x2=0; x<xsize;x++, x2+=bytesperpixel)  
             {     switch(format)
                   {     case CHARACTER: 
                                         fread(&cval,1,1,fp);
                                         rr = cval;
                                         putRGBbytes(x,y,ci,bpp,rr,rr,rr);
                                         break;
                         case WORD: 
                                         fread(&buf,2,1,fp);
                                         rr = ((buf[0]-128)<<8) + buf[1];
                                         putRGBbytes(x,y,ci,bpp,rr,rr,rr);
                                         break;
                         case INTEGER:   
                                         fread(&buf,4,1,fp);
                                         rr = ((buf[0]-128)<<8) + buf[1];
                                         putRGBbytes(x,y,ci,bpp,rr,rr,rr);
                                         break;
                         case FLOAT:    
                                         fread(&ival,4,1,fp);
           #ifdef LITTLE_ENDIAN
                                         swapbytes((uchar*)&ival,4,4);
           #endif
                                         fval = ieee754(ival);
                                         answer = min(65535, max(0,(int)(datamax*fval)));
                                         rr = answer;
                                         putRGBbytes(x,y,ci,bpp,rr,rr,rr);
                                         break; 
                         case DOUBLE: 
                                         fread(&lival,8,1,fp);
           #ifdef LITTLE_ENDIAN
                                         swapbytes((uchar*)&lival,8,8);
           #endif
                                         dval = ieee754d(lival);
                                         answer = min(65535, max(0,(int)(datamax*dval)));
                                         rr = answer;
                                         putRGBbytes(x,y,ci,bpp,rr,rr,rr);
                                         break; 
                   }
             }
             g.wantr=1;g.wantg=1;g.wantb=1;
       }             
  }      

  if(extend)
  {
      //// Read extension after image data in file
      //// Read more bytes to get to 2880 byte boundary
      //// Wish this was written down in the spec

      z[ci].metadata_extension = mcount;
      filepos = ftell(fp);
      fseek(fp, 2880-filepos%2880, SEEK_CUR);
       while(!feof(fp))
       {       fread(&record,2880,1,fp); 

               if(mcount > MAXRECORDS)
                     message("Too much metadata in FITS file\nExtra metadata will be ignored\nIncrease MAXRECORDS");
               z[ci].metadata[mcount] = new char[2880];
               strcpy(z[ci].metadata[mcount], record);
               mcount++;

               done = 0;
               for(k=0;k<36;k++)
               {    memcpy((void*)card, record+k*80, 80);
                    memcpy(keyword, card, 7);
                    if(!strcmp(keyword,"END    ")) done=1;
                    if(done) break;
               }
               if(done) break;
       }
  }       

  fclose(fp);
  z[ci].metadata_records = mcount;
  setSVGApalette(1); 
  return 0;
}  
          

//--------------------------------------------------------------------------//
// ieee754
// convert binary fraction to base 10 decimal
// for the abomination known as ieee754
//--------------------------------------------------------------------------//
long int rieee754d(double number)     // reverse ieee conversion
{
    long int answer;
    lieeedata data;
    data.output = number;
    answer = data.input;
    return answer;
}
int rieee754(float number)       // reverse ieee conversion
{
    int answer;
    ieeedata data;
    data.output = number;
    answer = data.input;
    return answer;
}
float ieee754(int number)       // 4 byte int to ieee754 float
{
    float answer;
    ieeedata data;
    data.input = number;
    answer = data.output;
    return answer;
}
double ieee754d(long int number)  // 8 byte int to ieee754 double
{
    double answer;
    lieeedata data;
    data.input = number;
    answer = data.output;
    return answer;
}

         

//--------------------------------------------------------------------------//
//  setupfitsfile                                                           //
//--------------------------------------------------------------------------//
void setupfitsfile(PromptStruct *ps)
{
      static Dialog *dialog;
      int j,k,error=0;
      drawselectbox(OFF);
      g.getout=0; 
      if(memorylessthan(16384)){  message(g.nomemory,ERROR); return; } 
      dialog = new Dialog;
      if(dialog==NULL){ message(g.nomemory); return; }
      dialog->helptopic=45;
      strcpy(dialog->title, "FITS file parameters");

      ////--Radio buttons--////

      strcpy(dialog->radio[0][0],"Data type");
      strcpy(dialog->radio[0][1],"8-bit unsigned integer  (char)");
      strcpy(dialog->radio[0][2],"16-bit unsigned integer (word)");
      strcpy(dialog->radio[0][3],"32-bit unsigned integer (int)");
      strcpy(dialog->radio[0][4],"32-bit IEEE 754 floating point");
      strcpy(dialog->radio[0][5],"64-bit IEEE 754 floating point");

      strcpy(dialog->radio[1][0],"Metadata");
      strcpy(dialog->radio[1][1],"Keep existing metadata");
      strcpy(dialog->radio[1][2],"Delete metadata");
      strcpy(dialog->radio[1][3],"Edit metadata");

      dialog->radiono[0]=6;
      dialog->radiono[1]=3;
      for(j=0;j<10;j++) for(k=0;k<20;k++) dialog->radiotype[j][k]=RADIO;
      dialog->noofradios=2;

      if(z[ci].subformat !=0)
      {     switch(z[ci].subformat)
            {     case CHARACTER: dialog->radioset[0] = 1; break;
                  case WORD     : dialog->radioset[0] = 2; break;
                  case INTEGER  : dialog->radioset[0] = 3; break;
                  case FLOAT    : dialog->radioset[0] = 4; break;
                  case DOUBLE   : dialog->radioset[0] = 5; break;
            }
      }else if(z[ci].colortype==GRAY)
      {     switch (z[ci].bpp)
            {     case  8: dialog->radioset[0] = 1; break;
                  case 16: dialog->radioset[0] = 2; break;
                  default: error++;
            } 
      }else
      {     switch (z[ci].bpp)
            {     case 24: dialog->radioset[0] = 1; break;
                  case 32: dialog->radioset[0] = 1; break;
                  case 48: dialog->radioset[0] = 2; break;
                  default: error++;
            } 
      }
      if(error)
      {    message("Image pixel depth is\nincompatible with FITS format", ERROR);
           return;
      }
      
      dialog->radioset[1] = 1;
    
      dialog->noofboxes=0;
      dialog->want_changecicb = 0;
      dialog->f1 = writefitsfilecheck;
      dialog->f2 = null;
      dialog->f3 = writefitsfilecancel;
      dialog->f4 = null;
      dialog->f5 = null;
      dialog->f6 = null;
      dialog->f7 = null;
      dialog->f8 = null;
      dialog->f9 = null;
      dialog->width = 0;  // calculate automatically
      dialog->height = 0; // calculate automatically
      dialog->transient = 1;
      dialog->wantraise = 0;
      dialog->radiousexy = 0;
      dialog->boxusexy = 0;
      dialog->radioextrawidth = 0;
      dialog->ptr[0] = ps;

      dialog->message[0]=0;      
      g.signif = 1;
      dialog->busy = 0;
      dialogbox(dialog);     
      return;
}


//-------------------------------------------------------------------------//
// writefitsfilecancel                                                     //
//-------------------------------------------------------------------------//
void writefitsfilecancel(dialoginfo *a)
{
   a=a;
}


//--------------------------------------------------------------------------//
//  writefitsfilecheck                                                      //
//--------------------------------------------------------------------------//
void writefitsfilecheck(dialoginfo *a, int radio, int box, int boxbutton)
{
  a=a;radio=radio;box=box;boxbutton=boxbutton;
  char record[2881];
  char tempstring[1024];
  char buf[4];
  int want_color = 1 - g.usegrayscale;
  int bpp, channel, errors=0, f, i, j, k, ino, xpos, ypos, xsize, ysize, xend, yend, 
      extra_bytes, v=0, rr, gg, bb, bitpix, naxis, recno, x, y;
  long int lval; 
  float    fval;
  double   dval, datamin, datamax;
  PromptStruct *ps = (PromptStruct*)a->ptr[0];
  char *filename  = ps->text;       // filename w/o .gz extension
  int format = a->radioset[0];

  if(radio != -2) return;   //// User clicked Ok or Enter

  ino = ci;
  if(a->radioset[1]==2)     //// What to do with metadata
  {      message("Deleting metadata");
         if(z[ino].metadata!=NULL)
         {    for(k=0; k<z[ino].metadata_records; k++)
                    delete[] z[ino].metadata[k]; 
              z[ino].metadata=NULL; 
              z[ino].metadata_records = 0;
              z[ino].metadata_extension = 0;
         }
  }
  if(a->radioset[1]==3)     //// What to do with metadata
      message("Editing metadata not yet implemented");
      
  FILE *fp;
  bpp = z[ino].bpp;
  xpos = z[ino].xpos;
  ypos = z[ino].ypos;
  xsize = z[ino].xsize;
  ysize = z[ino].ysize;
  xend = xpos + xsize - 1;
  yend = ypos + ysize - 1;
  f = z[ino].cf;
  datamin = 0.0;
  datamax = 65535.0;

  if((fp=open_file(filename, "w"))==NULL) return; 

  switch(format)
  {     
        case 1:  bitpix=8;   format=CHARACTER; break;
        case 2:  bitpix=16;  format=WORD; break;
        case 3:  bitpix=32;  format=INTEGER; break;
        case 4:  bitpix=-32; format=FLOAT; break;
        case 5:  bitpix=-64; format=DOUBLE; break;
        default:
             sprintf(tempstring,"Unsupported format in writefitscheck: %d",format);
             message(tempstring, ERROR);
             fclose(fp);
             return;
  }

  if(z[ino].colortype == GRAY)
  {    naxis  = 2;
       datamin = 65535.0;
       datamax = 0.0;
       for(y=0;y<ysize;y++)
       for(x=0;x<xsize;x++)
       {    //v = pixelat(z[ino].image[f][y]+x*z[ino].bpp, bpp);
            datamin = (double)min((uint)datamin, (uint)v);
            datamax = (double)max((uint)datamax, (uint)v);
       }
  }else       
  {    naxis  = 3;
       datamax = (double)g.maxred[bpp];  // must all be same, therefore no 16bit color
  }


  //// If no existing records, create one new one
  if(!z[ino].metadata_records) 
  {      z[ino].metadata = new char*[MAXRECORDS];
         z[ino].metadata[0] = new char[2880];
         z[ino].metadata_records = 1;         // only 1 record
         z[ino].metadata_extension = 0;       // no extension
         
         memset(record,32,2880); // Pad with spaces
         recno = 0;
         add_fits_key(record, recno++, "SIMPLE", "T");  
         add_fits_key(record, recno++, "BITPIX", bitpix);  
         add_fits_key(record, recno++, "NAXIS" , naxis);  
         add_fits_key(record, recno++, "NAXIS1", xend - xpos + 1);  
         add_fits_key(record, recno++, "NAXIS2", yend - ypos + 1);  
         if(z[ino].colortype == COLOR)
              add_fits_key(record, recno++, "NAXIS3", 3);  
         add_fits_key(record, recno++, "DATAMIN", 0);  
         add_fits_key(record, recno++, "DATAMAX", datamax);  
         add_fits_key(record, recno++, "BZERO", "0.0");  
         add_fits_key(record, recno++, "BSCALE", "1.0");  
         sprintf(tempstring, "%s %s", basefilename(g.appname), g.version);
         add_fits_key(record, recno++, "HISTORY", tempstring);  
         add_fits_key(record, recno++, "END", "");  
  }else  
  //// Substitute values into existing records
  {
         strcpy(record, z[ino].metadata[0]);
         substitute_fits_key(record, "SIMPLE ", "T");  
         substitute_fits_key(record, "BITPIX ", bitpix);  
         substitute_fits_key(record, "NAXIS  " , naxis);  
         substitute_fits_key(record, "NAXIS1 ", xend - xpos + 1);  
         substitute_fits_key(record, "NAXIS2 ", yend - ypos + 1);  
         if(z[ino].colortype == COLOR)
              substitute_fits_key(record, "NAXIS3 ", 3);  
         substitute_fits_key(record, "DATAMIN", datamin);  
         substitute_fits_key(record, "DATAMAX", datamax);  
  }
  
  strcpy(z[ino].metadata[0], record);
  fwrite(z[ino].metadata[0], 2880, 1, fp);  

  //// If there are more pre-existing records prior to extension, save them
  for(k=1; k<z[ino].metadata_records; k++)
       fwrite(z[ino].metadata[k], 2880, 1, fp);  

  if(z[ino].colortype == GRAY)
  {
       for(j=yend;j>=ypos;j--)  // Must be backwards
       {    for(i=xpos;i<=xend;i++)
            {   
                  v = readpixelonimage(i,j, bpp, ino);
                  v = convertpixel(v, bpp, g.want_bpp, want_color);

                  switch(format)
                  {    case CHARACTER: //6, char
                               putbyte(v, fp);  
                               break;
                       case WORD:      //7, 16-bit int
                               buf[0] = (v>>8)-128;
                               buf[1] = v & 0xf;
                               fwrite(&buf,2,1,fp);  
                               break;
                       case INTEGER:   //1, 32-bit int
                               buf[0] = (v>>8)-128;
                               buf[1] = v & 0xf;
                               buf[2] = 0;
                               buf[3] = 0;
                               fwrite(&buf,4,1,fp);  
                               break;
                       case FLOAT:     //2, 32 bit ieee754
                               fval = (float) v;
                               fval /= 65536.0;
                               v = rieee754(fval);
                               swapbytes((uchar*)&v,4,4);
                               fwrite(&v,4,1,fp);  
                               break;
                       case DOUBLE:    //4,  64 bit ieee754  
                               dval = (double) v;
                               dval /= 65536.0;
                               lval = rieee754d(dval);
                               swapbytes((uchar*)&lval,8,8);
                               fwrite(&lval,8,1,fp);  
                               break;
                       default:
                               errors++;
                               break;
                  } 

            }
       }     
  }else  // color
  {
       for(channel=0; channel<3; channel++)
       for(j=ysize-1; j>=0; j--)  // Must be backwards
       {    for(i=0; i<xsize; i++)
            {   
                  RGBat(z[ino].image[f][j-ypos]+(i-xpos)*g.off[bpp], bpp, rr, gg, bb);
                  switch(channel)
                  {    case 0: v = rr; break;  
                       case 1: v = gg; break;  
                       case 2: v = bb; break;  
                  }
                  switch(format)
                  {    case CHARACTER: //6, char
                               putbyte(v, fp);  
                               break;
                       case WORD:      //7, 16-bit int
                               buf[0] = (v>>8)-128;
                               buf[1] = v & 0xf;
                               fwrite(&buf,2,1,fp);  
                               break;
                       case INTEGER:   //1, 32-bit int
                               buf[0] = (v>>8)-128;
                               buf[1] = v & 0xf;
                               buf[2] = 0;
                               buf[3] = 0;
                               fwrite(&buf,4,1,fp);  
                               break;
                       case FLOAT:     //2, 32 bit ieee754
                               fval = (float) v;
                               fval /= 65536.0;
                               v = rieee754(fval);
                               swapbytes((uchar*)&v,4,4);
                               fwrite(&v,4,1,fp);  
                               break;
                       case DOUBLE:    //4,  64 bit ieee754  
                               dval = (double) v;
                               dval /= 65536.0;
                               lval = rieee754d(dval);
                               swapbytes((uchar*)&lval,8,8);
                               fwrite(&lval,8,1,fp);  
                               break;
                       default:
                               errors++;
                               break;
                  } 
            }
       }     
  }

  //// Pad with more spaces to next 2880 byte boundary
  extra_bytes = 2880 - g.off[g.want_bpp]*(xend-xpos+1)*(yend-ypos+1) % 2880;
  memset(record,32,2880); // Pad with spaces
  fwrite(record,extra_bytes,1,fp);  

  //// Save remaining extension metadata
  for(k=z[ino].metadata_records; k<z[ino].metadata_extension; k++)
       fwrite(z[ino].metadata[k], 2880, 1, fp);  


  close_file(fp);  
  save_image_file_part3(ps);
  if(errors) message("Error writing FITS File");
  return;
}

   

//--------------------------------------------------------------------------//
//  add_fits_key                                                            //
//--------------------------------------------------------------------------//
void add_fits_key(char *r, int n, const char *keyword, int value)
{  
   char tempstring[81];
   sprintf(tempstring,"%d",value);
   add_fits_key(r,n,keyword,tempstring);  
}
void add_fits_key(char *r, int n, const char *keyword, const char *value)
{ 
   int len = strlen(value);
   int offset = n*80;
   memset(r+offset,32,80);
   memcpy(r+offset, (void*)keyword, strlen(keyword));
   if(strlen(value)) 
   {     r[offset+8] = '=';
         memcpy(r+offset+30-len, (void*)value, strlen(value));
   }
}

//--------------------------------------------------------------------------//
//  substitute_fits_key                                                     //
//--------------------------------------------------------------------------//
void substitute_fits_key(char *r, const char *keyword, int value)
{  
   char tempstring[81];
   sprintf(tempstring,"%d",value);
   substitute_fits_key(r,keyword,tempstring);  
}
void substitute_fits_key(char *r, const char *keyword, const char *value)
{ 
   int k, n=0, len, offset;
   char testkeyword[8];
   testkeyword[7]=0;

   for(k=0;k<36;k++)
   {      memset(testkeyword, 32, 7);
          memcpy(testkeyword, r+k*80, 7);
          if(!strcmp(testkeyword, keyword)){ n=k; break; }
   }
   if(n)
   {
          len = strlen(value);
          offset = n*80;
          memset(r+offset,32,80);
          memcpy(r+offset, (void*)keyword, strlen(keyword));
          if(strlen(value)) 
          {     r[offset+8] = '=';
                memcpy(r+offset+30-len, (void*)value, strlen(value));
          }
   }
}
