//--------------------------------------------------------------------------//
// xmtnimage35.cc                                                           //
// Latest revision: 08-15-2023                                              //
// See xmtnimage.h for Copyright Notice                                     //
// brightness, contrast                                                     //
//--------------------------------------------------------------------------//

#include "xmtnimage.h"

extern Globals     g;
extern Image      *z;
extern int         ci;
extern PrinterStruct printer;
static char notcolor[] = "Cannot change RGB of grayscale image\nConvert image to color first"; 
static char notbw[] = "Cannot change grayscale of RGB image\nConvert image to grayscale first"; 

//-------------------------------------------------------------------------//
// changecontrast2                                                         //
//-------------------------------------------------------------------------//
void changecontrast2(int noofargs, char **arg)
{
   if(noofargs>0){ changecontrast(noofargs, arg); return; }
   int i,k;
   char title[128] = "Color";
   switchto(ci);
   selectimage(ci);
   g.getout = 0;

   //// keep static, used in changegraycontrastcb. copies are made in plot()
   static array<int> ydata(256,3);   
   static array<double> xdata(256);
   array<char> ytitle(256,3);
   for(k=0; k<3; k++) ytitle.p[k] = new char[256];
   for(i=0; i<256; i++)
   {   ydata.p[0][i] = z[ci].red_brightness[i]; 
       ydata.p[1][i] = z[ci].green_brightness[i]; 
       ydata.p[2][i] = z[ci].blue_brightness[i]; 
       xdata.data[i] = i;
   } 
   strcpy(ytitle.p[0]," Red");
   strcpy(ytitle.p[1]," Gren");
   strcpy(ytitle.p[2]," Blue");
   plot(title, 
       (char*)"Contrast", 
       ytitle.p, 
       xdata.data, 
       ydata.p, 
       256, 
       3, 
       BEZIER, 
       0, 
       255, 
       NULL, 
       null,             // f1
       null,             // f2
       image_bounds,     // f3
       null,             // f4
       changecontrastcb, // f5
       null,             // f6
       14);
}


//-------------------------------------------------------------------------//
// changegraycontrast                                                      //
//-------------------------------------------------------------------------//
void changegraycontrast(int noofargs, char **arg)
{
   if(noofargs>0){ changecontrast(noofargs, arg); return; }
   int i,k;
   char title[128] = "Grayscale";

   //// keep static, used in changegraycontrastcb. copies are made in plot()
   static array<int> ydata(256,1);   
   static array<double> xdata(256);
   array<char> ytitle(256,1);
   for(k=0; k<1; k++) ytitle.p[k] = new char[256];
   for(i=0; i<256; i++)
   {   ydata.p[0][i] = z[ci].gray_brightness[i]; 
       xdata.data[i] = i;
   } 
   strcpy(ytitle.p[0]," Grayscale");
   plot(title, 
       (char*)"Grayscale", 
       ytitle.p, 
       xdata.data, 
       ydata.p, 
       256, 
       1, 
       BEZIER, 
       0, 
       255, 
       NULL, 
       null,                 // f1
       null,                 // f2
       image_bounds,         // f3
       null,                 // f4
       changegraycontrastcb, // f5
       null,                 // f6
       14);
}


//-------------------------------------------------------------------------//
// change_intensity                                                        //
//-------------------------------------------------------------------------//
void change_intensity(int noofargs, char **arg)
{
   int k;
   if(noofargs>0)
   {    noofargs = 6;
        strcpy(arg[6], arg[1]);
        for(k=1;k<6;k++) strcpy(arg[k], "100");
        changecontrast(noofargs, arg); 
   }else message("Intensity increment not specified");
}
  


//--------------------------------------------------------------------------//
// image_stats                                                              //
//--------------------------------------------------------------------------//
void image_stats(int x1, int y1, int x2, int y2, int &ino, int &bpp, int &ymin, 
     int &ymax, int &rmin, int &gmin, int &bmin, int &rmax, int &gmax, int &bmax)
{
  int f,i,j,ii,jj,rr,gg,bb,value=0;
  ymax = rmax = gmax = bmax = 0; 
  ymin = rmin = gmin = bmin = MAXINT; 
  for(j=y1;j<=y2;j++)
  for(i=x1;i<=x2;i++)
  {   
     if(!g.selected_is_square && !inside_irregular_region(i,j)) continue;
     ino = whichimage(i, j, bpp);
     f = z[ino].cf;
     jj = j - z[ino].ypos;
     ii = g.off[bpp] * (i - z[ino].xpos);
     value = pixelat(z[ino].image[f][jj]+ii, bpp);
     valuetoRGB(value,rr,gg,bb,bpp);
     ymax = max(ymax, value);
     ymin = min(ymin, value);
     rmax = max(rmax, rr);
     rmin = min(rmin, rr);
     gmax = max(gmax, gg);
     gmin = min(gmin, gg);
     bmax = max(bmax, bb);
     bmin = min(bmin, bb);
  }
}


//--------------------------------------------------------------------------//
// image_bounds - get data for draw_graph                                   //
//--------------------------------------------------------------------------//
void image_bounds(void *p, int n1, int n2)
{
   n1=n1;n2=n2;
   PlotData *pd = (PlotData *)p;
   int x1,y1,x2,y2,ino,bpp,ymin,ymax,rmin,gmin,bmin,rmax,gmax,bmax;
   x1=g.ulx; x2=g.lrx; y1=g.uly; y2=g.lry;
   image_stats(x1,y1,x2,y2,ino,bpp,ymin,ymax,rmin,gmin,bmin,rmax,gmax,bmax);
   //// In anticipation of converting image to 24 bpp
   if(bpp==8){ bpp=24; rmin*=4; rmax=255; gmin*=4; gmax*=4; bmin*=4; bmax*=4; }
   switch(pd->focus)
   {    case 0:  pd->ymin=(double)rmin; 
                 pd->ymax=(double)rmax;
                 pd->ulimit=g.maxred[bpp];
                 break;
        case 1:  pd->ymin=(double)gmin; 
                 pd->ymax=(double)gmax;
                 pd->ulimit=g.maxgreen[bpp];
                 break;
        case 2:  pd->ymin=(double)bmin; 
                 pd->ymax=(double)bmax; 
                 pd->ulimit=g.maxblue[bpp];
                 break;
   }
}


//--------------------------------------------------------------------------//
// changecontrastcb                                                         //
// graphical, entire image only                                             //
//--------------------------------------------------------------------------//
void changecontrastcb(void *p, int n1, int n2)
{
  array<int> data48(65536,3);   
  n1 = n1; n2 = n2;
  int focus,hitgray=0,i,j,k,kk,ino,bpp,rr,gg,bb,value=0,rr2,gg2,bb2;
  int a,b,c,anytagged,oci;
  int oinmenu=g.inmenu;
  PlotData *pd = (PlotData*)p;
  int x1,x2,y1,y2;

  if(pd->ino<=0)
  {
     x1=g.ulx;
     x2=g.lrx;
     y1=g.uly;
     y2=g.lry;
  } else
  {
     switchto(pd->ino);
     selectimage(pd->ino);
     g.selectedimage = pd->ino;
     x1 = z[ci].xpos;
     y1 = z[ci].ypos;
     x2 = x1+z[ci].xsize;
     y2 = y1+z[ci].ysize;
     ino = ci;
  }

  g.inmenu = 0;
  printstatus(CALCULATING);
  focus = pd->focus;
  ino = g.selectedimage;
  printstatus(BUSY);

  static int hit = 0;
  static int ofocus = 0;
  static double olddata[256];
  int changed = 0;

  //// Make sure data is changed before starting time-consuming redraw
  if(!hit)
       for(k=0;k<256;k++) olddata[k]=0.0;
  for(k=0;k<256;k++)
       if(pd->data[focus][k] != olddata[k]){ changed++; break; }
  for(k=0;k<256;k++)
       olddata[k] = pd->data[focus][k];
  if(focus != ofocus) changed++;
  ofocus = focus;


  //// Return after first box drawn, nothing changed
  if(!changed ){ g.inmenu = oinmenu; return; } 
  hit = 1;
  if(g.selectedimage>0) restoreimage(0); // Don't restore colortype

  //// remap all the colors
  //// If in 48 bpp mode, do the interpolation

  anytagged = 0;
  for(k=1; k<g.image_count; k++)
     if(z[k].tagged==1){ anytagged=1; break;}
  oci = ci;

  for(kk=1; kk<g.image_count; kk++)
  {   if((z[kk].tagged==1) || (!anytagged && kk==oci))
      {    
          x1 = z[kk].xpos;
          y1 = z[kk].ypos;
          x2 = x1+z[kk].xsize-1;
          y2 = y1+z[kk].ysize-1;
          switchto(kk);
          ino = ci;
          bpp = z[ino].bpp;
          z[kk].hit = 1;

          if(z[kk].bpp==48)
          {
             for(i=0;i<256;i++)
             for(j=0;j<256;j++)
             {
                 a = (int)pd->data[0][i];
                 b = (int)pd->data[0][min(255,i+1)];
                 c = a*(256-j) + b*j;
                 data48.p[0][i*256+j] = c;

                 a = (int)pd->data[1][i];
                 b = (int)pd->data[1][min(255,i+1)];
                 c = a*(256-j) + b*j;
                 data48.p[1][i*256+j] = (int)pd->data[1][i] * 256;

                 a = (int)pd->data[2][i];
                 b = (int)pd->data[2][min(255,i+1)];
                 c = a*(256-j) + b*j;
                 data48.p[2][i*256+j] = (int)pd->data[2][i] * 256;
             } 
          }

          //// If in 8bpp mode, just change colormap
          if(g.selectedimage!=-1 && z[g.selectedimage].bpp==8)
          {  
              if(z[ino].colortype == GRAY) hitgray=1;
              else 
              {   for(k=0;k<256;k++)
                  {   z[ino].palette[k].red   = (int)pd->data[0][k]/4; 
                      z[ino].palette[k].green = (int)pd->data[1][k]/4; 
                      z[ino].palette[k].blue  = (int)pd->data[2][k]/4;
                  }
                  setpalette(z[ino].palette);
              }
          }else
          {
              for(j=y1; j<y2; j++)
              for(i=x1; i<x2; i++)
              {   
                 if(!g.selected_is_square && !inside_irregular_region(i,j)) continue;

                 //// Correct way of reading pixels, works for 48 bpp
                 ino = whichimage(i,j,bpp);
                 if(ino<0) continue;
                 if(z[ino].colortype==GRAY)
                       value = readpixelonimage(i,j,bpp,ino);
                 else             
                       readRGBonimage(i,j,bpp,ino,rr,gg,bb,-2);

                 if(z[ino].colortype == GRAY) { hitgray=1; continue; }

                 if(ino>=0 && z[ino].hit==0 && z[ino].colortype==INDEXED) 
                 {   z[ino].hit=2;    // Must be set before change_image_depth                   
                     if(change_image_depth(ino,24,0)!=OK) 
                     {   message("Insufficient image buffers available",ERROR);
                         g.getout=1; 
                         return;
                     }
                     bpp=24;

                 }

                 if(!z[ino].hit) z[ino].hit=1;

                 //// rr,gg,bb must be less than 256 to fit on graph
                 //// There can only be 256 r,g,b values
                 //// For 48 bit we have to interpolate, otherwise it loses information

                 if(bpp <= 32)
                 {
                     convertRGBpixel(rr,gg,bb,bpp,g.bitsperpixel);
                     rr2 = (int)pd->data[0][rr];
                     gg2 = (int)pd->data[1][gg];
                     bb2 = (int)pd->data[2][bb];
                     if(rr2 == rr && gg2 == gg && bb2 == bb) continue;

                     convertRGBpixel(rr2,gg2,bb2,g.bitsperpixel,bpp);
                     rr = max(0, min(rr2, (int)g.maxred[bpp]));
                     gg = max(0, min(gg2, (int)g.maxgreen[bpp]));
                     bb = max(0, min(bb2, (int)g.maxblue[bpp]));
                 }else
                 {
                     rr = data48.p[0][rr];
                     gg = data48.p[1][gg];
                     bb = data48.p[2][bb];         
                 }

                 //// Correct way of setting pixels, works for 48 bpp
                 if(z[ino].colortype==GRAY)
                      setpixelonimage(i,j,value,SET,bpp);
                 else
                      setRGBonimage(i,j,rr,gg,bb,SET,bpp,-1,1);

                 if(ino>=0 && (!z[ino].hit)) z[ino].hit=1;
              }
              for(k=0; k<g.image_count; k++) 
              {   if(z[k].hit==2)
                  {   z[k].hit=1;
                      change_image_depth(k,8,0);
                  }
              }
          }
      }
  }
  for(k=0;k<256;k++)
  {    z[ino].red_brightness[k]   = (int)pd->data[0][k];
       z[ino].green_brightness[k] = (int)pd->data[1][k];
       z[ino].blue_brightness[k]  = (int)pd->data[2][k];
  }
  if(hitgray) message(notcolor, WARNING);
  g.inmenu = oinmenu;  
  for(k=0; k<g.image_count; k++) 
  {  if(z[k].hit) 
     {   rebuild_display(k); redraw(k); z[k].hit=0; 
     }
  }
  printstatus(NORMAL);
}


//-------------------------------------------------------------------------//
// addvalue                                                                //
// Change color/intensity of selected region by adding a value.            //
//-------------------------------------------------------------------------//
void addvalue(int noofargs, char **arg)
{
    int dr=g.raddvalue, dg=g.gaddvalue, db=g.baddvalue;
    int darker=g.vaddvalue;
    if(memorylessthan(16384)){ message(g.nomemory,ERROR); return; }
    int k,bpp=0,maxbpp=0,gotbw=0,gotcolor=0,white,maxwhite=0,black,helptopic=14;

    static clickboxinfo* item;
    g.getout=0;
    char temp[128];
    int x1=g.ulx;
    int x2=g.lrx-1;
    int y1=g.uly;
    int y2=g.lry-1;

    for(k=0; k<g.image_count; k++) z[k].hit=0;
     
    pointinfo(x1,y1,bpp,gotcolor,gotbw,white,black);
    maxbpp=max(bpp,maxbpp);
    maxwhite=max(white,maxwhite);
    pointinfo(x1,y2,bpp,gotcolor,gotbw,white,black);
    maxbpp=max(bpp,maxbpp);
    maxwhite=max(white,maxwhite);
    pointinfo(x2,y1,bpp,gotcolor,gotbw,white,black);
    maxbpp=max(bpp,maxbpp);
    maxwhite=max(white,maxwhite);
    pointinfo(x2,y2,bpp,gotcolor,gotbw,white,black);
    maxbpp=max(bpp,maxbpp);
    maxwhite=max(white,maxwhite);
    pointinfo((x1+x2)/2,(y1+y2)/2,bpp,gotcolor,gotbw,white,black);
    maxbpp=max(bpp,maxbpp);
    maxwhite=max(white,maxwhite);


    if(noofargs)
    {  if(noofargs>=1) darker=atoi(arg[1]);
       if(noofargs>=2) dr=atoi(arg[2]);
       if(noofargs>=3) dg=atoi(arg[3]);
       if(noofargs>=4) db=atoi(arg[4]);
       clickboxinfo c[3];
       c[0].noofbuttons=3;
       c[0].answer = dr;
       c[1].answer = dg;
       c[2].answer = db;
       addvalueok(c);
    }else
    { 
       if(gotbw)
       {  strcpy(temp,"Change pixel value");
          if(gotcolor) strcat(temp," in mono. part");
//          clickbox(temp, 14, &darker, -maxwhite, maxwhite, null, 
//               addvalueok, null, NULL, NULL, helptopic);
          item = new clickboxinfo[3];
          item[0].title = new char[128];        
          item[1].title = new char[128];        
          item[2].title = new char[128];        
          item[0].wantpreview=0;
          item[0].autounmanage=0;
          strcpy(item[0].title,"Mono");
          item[0].startval = darker;
          item[0].minval = -g.maxvalue[maxbpp];
          item[0].maxval = g.maxvalue[maxbpp];
          item[0].type = VALSLIDER;
          item[0].wantdragcb = 0;
          item[0].answers = new int[100]; // Must be 100 for multiclickbox cb
          item[1].answers = new int[100]; // Must be 100 for multiclickbox cb
          item[2].answers = new int[100]; // Must be 100 for multiclickbox cb
          item[0].form = NULL;
          item[0].path = NULL;

          strcpy(temp,"Pixel value adjustment");
          if(gotbw) strcat(temp," in grayscale part");

          item[0].ino = ci;
          item[0].noofbuttons = 1;
          item[0].f1 = null;
          item[0].f2 = null;
          item[0].f3 = null;
          item[0].f4 = null;
          item[0].f5 = addvalueok;
          item[0].f6 = addvaluefinish;
          item[0].f7 = null;
          item[0].f8 = null;
          multiclickbox(temp, 1, item, null, helptopic);
       }
       if(gotcolor)
       {   
          item = new clickboxinfo[3];
          item[0].title = new char[128];        
          item[0].wantpreview=0;
          item[0].autounmanage=0;
          strcpy(item[0].title,"Red");
          item[0].startval = dr;
          item[0].minval = -g.maxred[maxbpp];
          item[0].maxval = g.maxred[maxbpp];
          item[0].type = VALSLIDER;
          item[0].wantdragcb = 0;
          item[0].answers = new int[100]; // Must be 100 for multiclickbox cb
          item[0].form = NULL;
          item[0].path = NULL;

          item[1].title = new char[128];        
          strcpy(item[1].title,"Green");
          item[1].startval = dg;
          item[1].minval = -g.maxgreen[maxbpp];
          item[1].maxval = g.maxgreen[maxbpp];
          item[1].type = VALSLIDER;
          item[1].wantdragcb = 0;
          item[1].form = NULL;
          item[1].path = NULL;

          item[2].title = new char[128];        
          strcpy(item[2].title,"Blue");
          item[2].startval = db;
          item[2].minval = -g.maxblue[maxbpp];
          item[2].maxval = g.maxblue[maxbpp];
          item[2].type = VALSLIDER;
          item[2].wantdragcb = 0;
          item[2].form = NULL;
          item[2].path = NULL;

          strcpy(temp,"Pixel value adjustment");
          if(gotbw) strcat(temp," in color part");

          item[0].ino = ci;
          item[0].noofbuttons = 3;
          item[0].f1 = null;
          item[0].f2 = null;
          item[0].f3 = null;
          item[0].f4 = null;
          item[0].f5 = addvalueok;
          item[0].f6 = addvaluefinish;
          item[0].f7 = null;
          item[0].f8 = null;
          multiclickbox(temp, 3, item, null, helptopic);
       }
    }
}


//-------------------------------------------------------------------------//
// addvalueok                                                              //
//-------------------------------------------------------------------------//
void addvalueok(clickboxinfo *c)
{
   int dr=0,dg=0,db=0,darker=0,ino,f,j,k,kk,i,i2,rr,gg,bb,bpp=0,obpp=0,
       value=0,oci,x1,x2,y1,y2,rmax,gmax,bmax,anytagged;
   if(c[0].noofbuttons >= 2) dg = c[1].answer;
   if(c[0].noofbuttons >= 3) db = c[2].answer;
   drawselectbox(OFF);
   printstatus(BUSY);
   g.getout = 0;
   printstatus(BUSY);

   if(c->noofbuttons==3)   //// changing color
   { 
      dr = c[0].answer;
      dg = c[1].answer;
      db = c[2].answer;
      g.raddvalue = dr;
      g.gaddvalue = dg;
      g.baddvalue = db;
   }
   if(c->noofbuttons==1)   //// changing gray
   {
      darker = c[0].answer;
      g.vaddvalue = darker;
   }

   if(g.selectedimage>=0)          //// entire image
   {   oci = ci;
       anytagged = 0;  
       for(k=1; k<g.image_count; k++)
       if(z[k].tagged==1){ anytagged=1; break;}
    
       for(kk=1; kk<g.image_count; kk++)
       {    
           if((z[kk].tagged==1) || (!anytagged && kk==oci))
           {    
                switchto(kk);
                ino = ci;
                bpp = z[ino].bpp;
                f = z[ino].cf;
                if(z[ino].colortype==GRAY)
                {
                    for(j=0; j<z[ino].ysize; j++)
                    for(i=0,i2=0; i<z[ino].xsize; i++,i2+=g.off[bpp])
                    {   
                        value = pixelat(z[ino].image[f][j]+i2,bpp) + darker;
                        value = max(0, min(value, g.maxvalue[bpp]));
                        putpixelbytes(z[ino].image[f][j]+i2, value, 1, bpp, 1);
                    }
                } //// end of add value gray
                else 
                {   
                    //// If part of an indexed image, must convert to color
                    //// then regenerate colormap
                    if(ino>=0 && z[ino].hit==0 && z[ino].colortype==INDEXED) 
                    {   z[ino].hit=2;    // Must be set before change_image_depth                   
                        if(change_image_depth(ino,24,0)!=OK) 
                         {   message("Insufficient image buffers available",ERROR);
                             g.getout=1; 
                             return;
                         }
                         bpp=24;
                    }
                    //// start of add value color 
                    rmax = g.maxred[bpp];
                    gmax = g.maxgreen[bpp];
                    bmax = g.maxblue[bpp];

                    for(j=0; j<z[ino].ysize; j++)
                    {
                       for(i=0,i2=0; i<z[ino].xsize; i++,i2+=g.off[bpp])
                       {   RGBat(z[ino].image[f][j]+i2, bpp, rr, gg, bb);
                           rr = rr + dr;
                           gg = gg + dg;
                           bb = bb + db;
                           rr = max(0,min(rr,rmax));
                           gg = max(0,min(gg,gmax));
                           bb = max(0,min(bb,bmax));
                           setRGBbytes(z[ino].image[f][j]+i2, rr, gg, bb, bpp);
                       }
                   }
                   if(z[k].hit==2)
                   {     z[k].hit=1;
                         change_image_depth(k,8,0);
                   }
                } //// end of add value color
                rebuild_display(kk);   // create new img buffer
                redraw(kk);
                z[kk].touched = 1;
           }
       }
   }else if(!g.getout)             //// selected area
   {  
      x1 = g.ulx;  
      y1 = g.uly;
      x2 = g.lrx;  
      y2 = g.lry;
      for(j=y1; j<=y2; j++)
      {  if(keyhit())if(getcharacter()==27)break;  
         for(i=x1; i<=x2; i++)
         {   
             if(!g.selected_is_square && !inside_irregular_region(i,j)) continue;

             //// Correct way of reading pixels, works for 48 bpp
             ino = whichimage(i,j,bpp);
             if(z[ino].colortype==GRAY)
                   value = readpixelonimage(i,j,bpp,ino);
             else             
                   readRGBonimage(i,j,bpp,ino,rr,gg,bb,-2);

             if(ino>=0)
             {  if(z[ino].hit==0 && z[ino].colortype==INDEXED) 
                {  z[ino].hit=2;  // Must be set before change_image_depth
                   if(change_image_depth(ino, 24, 0)!=OK) 
                   {  message("Insufficient image buffers available",ERROR);
                      g.getout=1; 
                      break;
                   }
                   rebuild_display(ino);   // create new img buffer
                   bpp = 24;
                 }
             }
                 
             ////  Crossing boundary to image with different bpp
             if(bpp!=obpp)       
             {  if((ino>=0)&&(bpp==8))   
                {  obpp=bpp; 
                   switch_palette(ino);
                }
             }   
             if(z[ino].hit==0) z[ino].hit=1;
             if((bpp==8)||(ino>=0 && z[ino].colortype==GRAY))
             {  value += darker;
                value = min((uint)g.maxvalue[bpp],(uint)max(0,value));
             }else
             {
                rr = max(0,min((int)g.maxred[bpp],( rr+dr )));
                gg = max(0,min((int)g.maxgreen[bpp], ( gg+dg )));
                bb = max(0,min((int)g.maxblue[bpp], ( bb+db )));
             }

             //// Correct way of setting pixels, works for 48 bpp
             if(z[ino].colortype==GRAY)
                  setpixelonimage(i,j,value,SET,bpp);
             else
                  setRGBonimage(i,j,rr,gg,bb,SET,bpp,-1,1);  //noscreen=1, faster
         }
         if(g.getout)break;
      }
           
      ////  Re-palettize any color images touched if in 8-bpp screen mode
      for(k=0; k<g.image_count; k++) 
      {  if(z[k].hit==2) 
         {     z[k].hit=1;
               change_image_depth(k,8,0);
         }
         if(z[k].hit) repair(k);
      }
      g.getout=0;
   }  
   printstatus(NORMAL);
}


//-------------------------------------------------------------------------//
// addvaluefinish                                                          //
//-------------------------------------------------------------------------//
void addvaluefinish(clickboxinfo *c)
{
  delete[] c[0].answers;
  delete[] c[0].title;
  delete[] c[1].title;
  delete[] c[2].title;
  delete[] c;                                   
}


//--------------------------------------------------------------------------//
// addpalettecolors - add r,g,b,& i in palette by dr,dg,db,di               //
// d[] is value to add to 0=r 1=g 2=b 3=h 4=s 5=v.                          //
// If value[6] is 1, the change is permanent, otherwise it is temporary.    //
// Compatible with multiclickbox                                            //
// Caller sets taggingok to 0 if it is already looping through tagged images//
// otherwise only works in ino                                              //
//--------------------------------------------------------------------------//
void addpalettecolors(int d[100], int ino, int taggingok)
{ 
   int hitgray=0,f=0,i,i2,j,k,bpp,value,rr,gg,bb,obpp=0,xx,yy;
   int x1,y1,x2,y2,xpos,ypos;
   int dr,dg,db,dfac=1;
   int dh,ds,dv,di,hh,ss,vv;
   printstatus(BUSY);
   taggingok = taggingok;

   ////  If color, changing 'intensity' should change rg&b
   dr = d[0];
   dg = d[1];
   db = d[2];
   dh = d[3];
   ds = d[4];
   dv = d[5];
   di = d[6];
   d[7]=1;

   z[ino].hit = 1;
   if(g.selectedimage>=0 && ino>=0) // Entire image
   {   bpp = z[ino].bpp;
       f = z[ino].cf;
       if(z[ino].colortype==GRAY)
       {  
           hitgray = 1;
           for(j=0; j<z[ino].ysize; j++)
           for(i=0,i2=0; i<z[ino].xsize; i++,i2+=g.off[bpp])
           {   
               value = pixelat(z[ino].image[f][j]+i2,bpp);
               value = min((int)g.maxvalue[bpp], (int)max(0,(di + value)));
               putpixelbytes(z[ino].image[f][j]+i2, value, 1, bpp, 1);
               ////  This only needs to be done for grayscale
               if(z[ino].floatexists || z[ino].waveletexists)
               {   xx = i + z[ino].xpos;
                   yy = j + z[ino].ypos;
                   if(z[ino].floatexists) setfloatpixel(xx,yy,value,ino);
                   if(z[ino].waveletexists) setwaveletpixel(xx,yy,value,ino);
               }
           }
       }else
       {   //// If part of an indexed image, must convert to color
           //// then regenerate colormap

           if(ino>=0 && z[ino].hit==0 && z[ino].colortype==INDEXED) 
           {    z[ino].hit=2;    // Must be set before change_image_depth
                dfac = 4;
                if(change_image_depth(ino,24,0)!=OK) 
                {   message("Insufficient image buffers available",ERROR);
                    g.getout=1; 
                    return;
                }
                bpp=24;
           }

           for(j=0;j<z[ino].ysize;j++)
           for(i=0,i2=0;i<z[ino].xsize;i++,i2+=g.off[bpp])
           { 
               RGBat(z[ino].image[f][j]+i2, bpp, rr, gg, bb);
               rr = max(0,min(g.maxred[bpp],   rr + dr*dfac));
               gg = max(0,min(g.maxgreen[bpp], gg + dg*dfac));
               bb = max(0,min(g.maxblue[bpp],  bb + db*dfac));
               if(dh || ds || dv)
               {    RGBtoHSV(rr, gg, bb, hh, ss, vv, bpp);
                    hh = max(0,min(360, hh + dh));
                    ss = max(0,min(g.maxgray[bpp],  ss + ds));
                    vv = max(0,min(g.maxgray[bpp],  vv + dv*dfac));
                    HSVtoRGB(hh, ss, vv, rr, gg, bb, bpp);
               }
               setRGBbytes(z[ino].image[f][j]+i2, rr, gg, bb, bpp);
           }
           for(k=0; k<g.image_count; k++) 
           {   if(z[k].hit==2)
               {     z[k].hit=1;
                     change_image_depth(k,8,0);
               }
           }
       }
   }else                 // Selected region only
   {
       x1 = g.ulx;
       y1 = g.uly;
       x2 = g.lrx;
       y2 = g.lry;
       for(j=y1;j<=y2;j++)
       for(i=x1;i<=x2;i++)
       {   
             if(!g.selected_is_square && !inside_irregular_region(i,j)) continue;
             ino = whichimage(i,j,bpp);
             xpos = z[ino].xpos;
             ypos = z[ino].ypos;
             ////  Crossing boundary to image with different bpp

             if(ino>=0 && bpp!=obpp && bpp==8)
             {    obpp=bpp; 
                  switch_palette(ino);
             }   

             //// If part of an indexed image, must convert to color
             //// then regenerate colormap

             if(ino>=0 && z[ino].hit==0 && z[ino].colortype==INDEXED) 
             {    
                  z[ino].hit=2;    // Must be set before change_image_depth
                  if(change_image_depth(ino,24,0)==OK) bpp=24;
                  else
                  {   message("Insufficient image buffers available",ERROR);
                      g.getout=1; 
                      break;
                  }
                  rebuild_display(ino);
             }

             ////  Must convert to screen bits/pixel so the increment is the
             ////  same as the preview above. Otherwise the increase in brightness
             ////  could be doubled if, for example, a 16 bpp image was adjusted 
             ////  at 24 bpp screen depth.

             if(ino>=0 && z[ino].colortype==GRAY)
             {  
                  value = readpixelonimage(i,j,bpp,ino);
                  value = min((int)g.maxvalue[bpp], (int)(value+di));
                  hitgray = 1;
                  setpixelonimage(i,j,value,SET);
             }else
             {   
                  if(!between(i-xpos,0,z[ino].xsize-1) || !between(j-ypos,0,z[ino].ysize-1)) continue;
                  RGBat(z[ino].image[f][j-ypos]+(i-xpos)*g.off[bpp], bpp, rr, gg, bb);
                  rr = max(0,min((int)g.maxred[bpp], (int)(rr+dr)));
                  gg = max(0,min((int)g.maxgreen[bpp], (int)(gg+dg)));
                  bb = max(0,min((int)g.maxblue[bpp], (int)(bb+db)));
                  if(dh || ds || dv)
                  {    RGBtoHSV(rr, gg, bb, hh, ss, vv, bpp);
                       hh = max(0,min(360, hh+dh));
                       ss = max(0,min(g.maxgray[bpp],  ss+ds));
                       vv = max(0,min(g.maxgray[bpp],  vv+dv));
                       HSVtoRGB(hh, ss, vv, rr, gg, bb, bpp);
                  }
                  setRGBbytes(z[ino].image[f][j-ypos]+(i-xpos)*g.off[bpp], rr, gg, bb, bpp);
             }
       }            
       ////  Re-palettize any color images touched if in 8-bpp screen mode.
       ////  See invertcolors() in xmtnimage2.cc for an explanation of this code.
       for(k=0; k<g.image_count; k++) 
       {     if(z[k].hit==2) 
             {     z[k].hit=1;
                   change_image_depth(k,8,0);
             }
       }
   }
    
   if(hitgray && (d[0] || d[1] || d[2])) message(notcolor, WARNING);
   printstatus(NORMAL);
   return;
}


//--------------------------------------------------------------------------//
// changegraycontrastcb                                                     //
//--------------------------------------------------------------------------//
void changegraycontrastcb(void *p, int n1, int n2)
{
  n1=n1; n2=n2;
  int i,j,k,kk,ino,bpp,value,index,grayfac,anytagged,oci;
  double grayfacinv;
  int oinmenu=g.inmenu;
  PlotData *pd = (PlotData*)p;
  int x1=g.ulx;
  int x2=g.lrx;
  int y1=g.uly;
  int y2=g.lry;
  g.inmenu = 0;
  printstatus(CALCULATING);

  static int hit = 0;
  static double olddata[256];
  int changed = 0;

  switchto(pd->ino);

  //// Make sure data is changed before starting time-consuming redraw
  if(!hit)
       for(k=0;k<256;k++) olddata[k]=0.0;
  for(k=0;k<256;k++)
       if(pd->data[0][k] != olddata[k]){ changed++; break; }
  for(k=0;k<256;k++)
       olddata[k] = pd->data[0][k];
  hit = 1;

  //// Return after first box drawn, nothing changed
  if(!changed){ g.inmenu = oinmenu; return; } 
  if(g.selectedimage>0) restoreimage(0); // Don't restore colortype

  anytagged = 0;
  for(k=1; k<g.image_count; k++)
     if(z[k].tagged==1){ anytagged=1; break;}
  oci = ci;
  if(oci == 0) { message("No image selected"); return; }
  //// Start remapping the gray values according to the graph
  for(kk=1; kk<g.image_count; kk++)
  {  
      switchto(kk);
      if((z[kk].tagged==1) || (!anytagged && kk==oci))
      {    
          switchto(kk);
          ino = ci;
          x1 = z[kk].xpos;
          y1 = z[kk].ypos;
          x2 = x1+z[kk].xsize-1;
          y2 = y1+z[kk].ysize-1;
          bpp = z[ino].bpp;
          z[kk].hit = 1;

          ino = kk;
          if(!between(ino, 1, g.image_count)){ g.inmenu = oinmenu; return; }
          bpp = z[ino].bpp;
          grayfac = int(+g.maxvalue[bpp] / (double)255);
          grayfacinv = 1.0 / grayfac;     //// scale pixel value to 0-255
          if(z[ino].colortype != GRAY){ message(notbw, WARNING); g.inmenu = oinmenu; return; }
          printstatus(BUSY);

          //// If in 8bpp mode, just change colormap
          for(j=y1;j<=y2;j++)
          for(i=x1;i<=x2;i++)
          {      if(!g.selected_is_square && !inside_irregular_region(i,j)) continue;
                 value = readpixelonimage(i,j,bpp,ino);
                 if(ino < 0) continue;
                 index = int(value * grayfacinv);
                 if(z[ino].gray_brightness[index] == (int)pd->data[0][index]) continue;
                 value = grayfac * (int)pd->data[0][index];
                 value = max(0,min(value,(int)g.maxvalue[bpp]));
                 setpixelonimage(i,j,value,SET,z[ino].bpp);
          }
      }
      for(k=0;k<256;k++)
        z[kk].gray_brightness[k] = (int)pd->data[0][k];
  }
  g.inmenu = oinmenu;  
  for(k=0; k<g.image_count; k++) 
  {
      if(z[k].hit) 
      {  rebuild_display(k); 
         redraw(k);
      }
  }
  printstatus(NORMAL);
}

